@@ -142,7 +142,7 @@ panel.series <- function(x, y, recon, ...) {
   idx <- seq_along(groups);
   d <- data.frame(A = idx, B = idx);
 
-  r <- reconstruct(x, groups = groups);
+  r <- reconstruct(x, groups = groups, drop = FALSE);
 
   # Provide convenient defaults
   dots <- .defaults(dots, "type", "l");
@@ -238,7 +238,7 @@ plot.1d.ssa.reconstruction <- function(x, ...,
   colnames(m) <- mnames
 
   # Plot'em'all!
-  if (identical(plot.method, "matplot"))
+  if (identical(plot.method, "matplot") || !is.object(m))
     do.call(matplot, c(list(x = m), dots))
   else if (identical(plot.method, "native"))
     do.call(plot, c(list(m), dots))
