@@ -32,16 +32,17 @@
 #include <R_ext/Applic.h>
 #endif
 
+#include "fft_plan.h"
+
 typedef struct {
 #if HAVE_FFTW3_H
   fftw_complex * circ_freq;
-  fftw_plan r2c_plan;
-  fftw_plan c2r_plan;
 #else
   complex double * circ_freq;
 #endif
   R_len_t window;
   R_len_t length;
+  fft_plan *fft_plan;
 } hankel_matrix;
 
 static unsigned hankel_nrow(const void *matrix) {
@@ -55,17 +56,13 @@ static unsigned hankel_ncol(const void *matrix) {
 }
 
 #if HAVE_FFTW3_H
-static void free_circulant(hankel_matrix *h) {
-  fftw_free(h->circ_freq);
-  fftw_destroy_plan(h->r2c_plan);
-  fftw_destroy_plan(h->c2r_plan);
+static void free_plan(fft_plan *f) {
+  fftw_destroy_plan(f->r2c_plan);
+  fftw_destroy_plan(f->c2r_plan);
 }
 
-static void initialize_circulant(hankel_matrix *h,
-                                 const double *F, R_len_t N, R_len_t L) {
-  R_len_t K = N - L + 1, i;
+static void initialize_plan(fft_plan *f, R_len_t N) {
   fftw_complex *ocirc;
-  fftw_plan p1, p2;
   double *circ;
 
   /* Allocate needed memory */
@@ -73,8 +70,31 @@ static void initialize_circulant(hankel_matrix *h,
   ocirc = (fftw_complex*) fftw_malloc((N/2 + 1) * sizeof(fftw_complex));
 
   /* Estimate the best plans for given input length */
-  p1 = fftw_plan_dft_r2c_1d(N, circ, ocirc, FFTW_ESTIMATE);
-  p2 = fftw_plan_dft_c2r_1d(N, ocirc, circ, FFTW_ESTIMATE);
+  f->r2c_plan = fftw_plan_dft_r2c_1d(N, circ, ocirc, FFTW_ESTIMATE);
+  f->c2r_plan = fftw_plan_dft_c2r_1d(N, ocirc, circ, FFTW_ESTIMATE);
+
+  f->N = N;
+
+  fftw_free(circ);
+  fftw_free(ocirc);
+}
+
+static void free_circulant(hankel_matrix *h) {
+  fftw_free(h->circ_freq);
+}
+
+static void initialize_circulant(hankel_matrix *h, fft_plan *f,
+                                 const double *F, R_len_t N, R_len_t L) {
+  R_len_t K = N - L + 1, i;
+  fftw_complex *ocirc;
+  double *circ;
+
+  if (!valid_plan(f, N))
+    error("invalid FFT plan for given FFT length");
+
+  /* Allocate needed memory */
+  circ = (double*) fftw_malloc(N * sizeof(double));
+  ocirc = (fftw_complex*) fftw_malloc((N/2 + 1) * sizeof(fftw_complex));
 
   /* Fill input buffer */
   for (i = K-1; i < N; ++i)
@@ -85,22 +105,22 @@ static void initialize_circulant(hankel_matrix *h,
   }
 
   /* Run the plan on input data */
-  fftw_execute(p1);
+  fftw_execute_dft_r2c(f->r2c_plan, circ, ocirc);
 
   /* Cleanup and return */
   fftw_free(circ);
 
   h->circ_freq = ocirc;
-  h->r2c_plan = p1;
-  h->c2r_plan = p2;
   h->window = L;
   h->length = N;
+  h->fft_plan = f;
 }
 
 static void hankel_matmul(double* out,
                           const double* v,
                           const void* matrix) {
   const hankel_matrix *h = matrix;
+  const fft_plan *f = h->fft_plan;
   R_len_t N = h->length, L = h->window;
   R_len_t K = N - L + 1, i;
   double *circ;
@@ -116,14 +136,14 @@ static void hankel_matmul(double* out,
   memset(circ + K, 0, (L - 1)*sizeof(double));
 
   /* Compute the FFT of the reversed vector v */
-  fftw_execute_dft_r2c(h->r2c_plan, circ, ocirc);
+  fftw_execute_dft_r2c(f->r2c_plan, circ, ocirc);
 
   /* Dot-multiply with pre-computed FFT of toeplitz circulant */
   for (i = 0; i < (N/2 + 1); ++i)
     ocirc[i] = ocirc[i] * h->circ_freq[i];
 
   /* Compute the reverse transform to obtain result */
-  fftw_execute_dft_c2r(h->c2r_plan, ocirc, circ);
+  fftw_execute_dft_c2r(f->c2r_plan, ocirc, circ);
 
   /* Cleanup and return */
   for (i = 0; i < L; ++i)
@@ -137,6 +157,7 @@ static void hankel_tmatmul(double* out,
                            const double* v,
                            const void* matrix) {
   const hankel_matrix *h = matrix;
+  const fft_plan *f = h->fft_plan;
   R_len_t N = h->length, L = h->window;
   R_len_t K = N - L + 1, i;
   double *circ;
@@ -152,14 +173,14 @@ static void hankel_tmatmul(double* out,
     circ[i + K - 1] = v[L - i - 1];
 
   /* Compute the FFT of the reversed vector v */
-  fftw_execute_dft_r2c(h->r2c_plan, circ, ocirc);
+  fftw_execute_dft_r2c(f->r2c_plan, circ, ocirc);
 
   /* Dot-multiply with pre-computed FFT of toeplitz circulant */
   for (i = 0; i < (N/2 + 1); ++i)
     ocirc[i] = ocirc[i] * h->circ_freq[i];
 
   /* Compute the reverse transform to obtain result */
-  fftw_execute_dft_c2r(h->c2r_plan, ocirc, circ);
+  fftw_execute_dft_c2r(f->c2r_plan, ocirc, circ);
 
   /* Cleanup and return */
   for (i = 0; i < K; ++i)
@@ -171,13 +192,16 @@ static void hankel_tmatmul(double* out,
 
 static R_INLINE void hankelize_fft(double *F,
                                    const double *U, const double *V,
-                                   const hankel_matrix *h) {
-  R_len_t N = h->length, L = h->window;
-  R_len_t K = N - L + 1;
+                                   R_len_t L, R_len_t K,
+                                   const fft_plan *f) {
+  R_len_t N = K + L - 1;
   R_len_t i;
   double *iU, *iV;
   fftw_complex *cU, *cV;
 
+  if (!valid_plan(f, N))
+    error("invalid FFT plan for given FFT length");
+
   /* Allocate needed memory */
   iU = (double*) fftw_malloc(N * sizeof(double));
   iV = (double*) fftw_malloc(N * sizeof(double));
@@ -186,21 +210,21 @@ static R_INLINE void hankelize_fft(double *F,
 
   /* Fill in buffers */
   memcpy(iU, U, L*sizeof(double));
-  memset(iU+L, 0, (K - 1)*sizeof(double));
+  memset(iU + L, 0, (K - 1)*sizeof(double));
 
   memcpy(iV, V, K*sizeof(double));
-  memset(iV+K, 0, (L - 1)*sizeof(double));
+  memset(iV + K, 0, (L - 1)*sizeof(double));
 
   /* Compute the FFTs */
-  fftw_execute_dft_r2c(h->r2c_plan, iU, cU);
-  fftw_execute_dft_r2c(h->r2c_plan, iV, cV);
+  fftw_execute_dft_r2c(f->r2c_plan, iU, cU);
+  fftw_execute_dft_r2c(f->r2c_plan, iV, cV);
 
   /* Dot-multiply */
   for (i = 0; i < N/2 + 1; ++i)
     cU[i] = cU[i] * cV[i];
 
   /* Compute the inverse FFT */
-  fftw_execute_dft_c2r(h->c2r_plan, cU, iU);
+  fftw_execute_dft_c2r(f->c2r_plan, cU, iU);
 
   /* Form the result */
   for (i = 0; i < N; ++i) {
@@ -219,22 +243,79 @@ static R_INLINE void hankelize_fft(double *F,
   fftw_free(cU);
   fftw_free(cV);
 }
+
+static void compute_L_covariation_matrix_first_row(const double *F, R_len_t N, R_len_t L,
+                                                   double *R,
+                                                   const fft_plan *f) {
+  double *oF;
+  fftw_complex *iF, *iFc;
+  R_len_t i, K;
+
+  /* Length's check */
+  if ((L >= N) || (L < 1))
+    error("must be 'N' > 'L' >= 1");
+
+  K = N - L + 1;
+
+  /* Allocate needed memory */
+
+  oF = (double*) fftw_malloc(N * sizeof(double));
+  iF = (fftw_complex*) fftw_malloc((N/2 + 1) * sizeof(fftw_complex));
+  iFc = (fftw_complex*) fftw_malloc((N/2 + 1) * sizeof(fftw_complex));
+
+  memcpy(oF, F, N*sizeof(double));
+
+  fftw_execute_dft_r2c(f->r2c_plan, oF, iF);
+
+  memcpy(oF, F, K*sizeof(double));
+  memset(oF + K, 0, (N - K)*sizeof(double));
+
+  fftw_execute_dft_r2c(f->r2c_plan, oF, iFc);
+
+  /* Dot-product */
+  for (i = 0; i < N/2 + 1; ++i)
+    iF[i] = iF[i] * conj(iFc[i]);
+
+  fftw_execute_dft_c2r(f->c2r_plan, iF, oF);
+
+  /* Return */
+  for (i = 0; i < L; ++i)
+    R[i] = oF[i] / N;
+
+  /* Cleanup */
+  fftw_free(oF);
+  fftw_free(iF);
+  fftw_free(iFc);
+}
 #else
+static void free_plan(fft_plan *f) {
+  (void)f;
+}
+
+static void initialize_plan(fft_plan *f, R_len_t N) {
+  f->N = N;
+}
+
 static void free_circulant(hankel_matrix *h) {
   Free(h->circ_freq);
 }
 
-static void initialize_circulant(hankel_matrix *h,
+static void initialize_circulant(hankel_matrix *h, fft_plan *f,
                                  const double *F, R_len_t N, R_len_t L) {
   R_len_t K = N - L + 1, i;
   int *iwork, maxf, maxp;
   double *work;
   complex double * circ;
 
+  if (!valid_plan(f, N))
+    error("invalid FFT plan for given FFT length");
+
   /* Allocate needed memory */
   circ = Calloc(N, complex double);
 
   /* Estimate the best plans for given input length */
+  /* No, we cannot move this to plan object, because fft_factor
+   * has side effects :( */
   fft_factor(N, &maxf, &maxp);
   if (maxf == 0)
     error("fft factorization error");
@@ -260,6 +341,7 @@ static void initialize_circulant(hankel_matrix *h,
   h->circ_freq = circ;
   h->window = L;
   h->length = N;
+  h->fft_plan = f;
 }
 
 static void hankel_matmul(double* out,
@@ -352,14 +434,17 @@ static void hankel_tmatmul(double* out,
 
 static R_INLINE void hankelize_fft(double *F,
                                    const double *U, const double *V,
-                                   const hankel_matrix *h) {
-  R_len_t N = h->length, L = h->window;
-  R_len_t K = N - L + 1;
+                                   R_len_t L, R_len_t K,
+                                   const fft_plan *f) {
+  R_len_t N = K + L - 1;
   R_len_t i;
   int maxf, maxp, *iwork;
   double *work;
   complex double *iU, *iV;
 
+  if (!valid_plan(f, N))
+    error("invalid FFT plan for given FFT length");
+
   /* Estimate the best plans for given input length */
   fft_factor(N, &maxf, &maxp);
   if (maxf == 0)
@@ -409,6 +494,65 @@ static R_INLINE void hankelize_fft(double *F,
   Free(work);
   Free(iwork);
 }
+
+static void compute_L_covariation_matrix_first_row(const double *F, R_len_t N, R_len_t L,
+                                                   double *R,
+                                                   const fft_plan *f) {
+  R_len_t K = N - L + 1;
+  R_len_t i;
+  int maxf, maxp, *iwork;
+  double *work;
+  complex double *iF, *iFc;
+
+  if (!valid_plan(f, N))
+    error("invalid FFT plan for given FFT length");
+
+  /* Length's check */
+  if ((L >= N) || (L < 1))
+    error("must be 'N' > 'L' >= 1");
+
+  /* Estimate the best plans for given input length */
+  fft_factor(N, &maxf, &maxp);
+  if (maxf == 0)
+    error("fft factorization error");
+
+  /* Allocate needed memory */
+  iF = Calloc(N, complex double);
+  iFc = Calloc(N, complex double);
+  work = Calloc(4 * maxf, double);
+  iwork = Calloc(maxp, int);
+
+  /* Fill in buffers */
+  for (i = 0; i < N; ++i)
+    iF[i] = F[i];
+
+  for (i = 0; i < K; ++i)
+    iFc[i] = F[i];
+  memset(iFc + K, 0, (N - K)*sizeof(complex double));
+
+  /* Compute the FFTs */
+  fft_factor(N, &maxf, &maxp);
+  fft_work((double*)iF, ((double*)iF)+1, 1, N, 1, -2, work, iwork);
+  fft_factor(N, &maxf, &maxp);
+  fft_work((double*)iFc, ((double*)iFc)+1, 1, N, 1, -2, work, iwork);
+
+  /* Dot-product */
+  for (i = 0; i < N; ++i)
+    iF[i] = iF[i] * conj(iFc[i]);
+
+  /* Compute the inverse FFT */
+  fft_factor(N, &maxf, &maxp);
+  fft_work((double*)iF, ((double*)iF)+1, 1, N, 1, +2, work, iwork);
+
+  /* Form the result */
+  for (i = 0; i < L; ++i)
+    R[i] = creal(iF[i]) / N;
+
+  Free(iF);
+  Free(iFc);
+  Free(work);
+  Free(iwork);
+}
 #endif
 
 /* This is just direct R-to-C translation  */
@@ -446,6 +590,44 @@ static R_INLINE void hankelize(double *F,
   }
 }
 
+
+static void fft_plan_finalizer(SEXP ptr) {
+  fft_plan *f;
+
+  if (TYPEOF(ptr) != EXTPTRSXP)
+    return;
+
+  f = R_ExternalPtrAddr(ptr);
+  if (!f)
+    return;
+
+  free_plan(f);
+  Free(f);
+
+  R_ClearExternalPtr(ptr);
+}
+
+SEXP initialize_fft_plan(SEXP rN) {
+  R_len_t N;
+  fft_plan *f;
+  SEXP res;
+
+  N = INTEGER(rN)[0];
+
+  /* Allocate memory */
+  f = Calloc(1, fft_plan);
+
+  /* Do actual plan initialization */
+  initialize_plan(f, N);
+
+  /* Make an external pointer envelope */
+  PROTECT(res = R_MakeExternalPtr(f, install("fft plan"), R_NilValue));
+  R_RegisterCFinalizer(res, fft_plan_finalizer);
+  UNPROTECT(1);
+
+  return res;
+}
+
 static void hmat_finalizer(SEXP ptr) {
   ext_matrix *e;
   hankel_matrix *h;
@@ -469,7 +651,7 @@ static void hmat_finalizer(SEXP ptr) {
   R_ClearExternalPtr(ptr);
 }
 
-SEXP initialize_hmat(SEXP F, SEXP window) {
+SEXP initialize_hmat(SEXP F, SEXP window, SEXP fft_plan) {
   R_len_t N, L;
   hankel_matrix *h;
   ext_matrix *e;
@@ -488,11 +670,12 @@ SEXP initialize_hmat(SEXP F, SEXP window) {
 
   /* Build toeplitz circulants for hankel matrix */
   h = Calloc(1, hankel_matrix);
-  initialize_circulant(h, REAL(F), N, L);
+
+  initialize_circulant(h, R_ExternalPtrAddr(fft_plan), REAL(F), N, L);
   e->matrix = h;
 
   /* Make an external pointer envelope */
-  hmat = R_MakeExternalPtr(e, install("external matrix"), R_NilValue);
+  hmat = R_MakeExternalPtr(e, install("external matrix"), fft_plan);
   R_RegisterCFinalizer(hmat, hmat_finalizer);
 
   return hmat;
@@ -627,37 +810,63 @@ SEXP hankelize_one(SEXP U, SEXP V) {
   return F;
 }
 
-SEXP hankelize_one_fft(SEXP U, SEXP V, SEXP hmat) {
-  SEXP F = NILSXP, tchk;
+SEXP hankelize_one_fft(SEXP U, SEXP V, SEXP fftplan) {
+  SEXP F = NILSXP;
+  fft_plan *plan;
+  double *rU = REAL(U), *rV = REAL(V), *rF;
+  R_len_t L, K, N;
 
-  /* Perform a type checking */
-  PROTECT(tchk = is_hmat(hmat));
+  L = length(U); K = length(V);
+  N = K + L - 1;
 
-  if (LOGICAL(tchk)[0]) {
-    ext_matrix *e;
-    hankel_matrix *h;
-    double *rU = REAL(U), *rV = REAL(V), *rF;
+  /* Grab needed data */
+  plan = R_ExternalPtrAddr(fftplan);
 
-    /* Grab needed data */
-    e = R_ExternalPtrAddr(hmat);
-    h = e->matrix;
+  /* Allocate buffer for output */
+  PROTECT(F = allocVector(REALSXP, N));
+  rF = REAL(F);
 
-    /* Allocate buffer for output */
-    PROTECT(F = allocVector(REALSXP, h->length));
-    rF = REAL(F);
-
-    /* Perform the actual hankelization */
-    hankelize_fft(rF, rU, rV, h);
-
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel matrix");
+  /* Perform the actual hankelization */
+  hankelize_fft(rF, rU, rV, L, K, plan);
 
   UNPROTECT(1);
 
   return F;
 }
 
+SEXP hankelize_multi_fft(SEXP U, SEXP V, SEXP fftplan) {
+  double *rU = REAL(U), *rV = REAL(V), *rF;
+  R_len_t L, K, N, i, count;
+  SEXP F = NILSXP;
+  fft_plan *plan;
+  int *dimu, *dimv;
+
+  /* Calculate length of inputs and output */
+  dimu = INTEGER(getAttrib(U, R_DimSymbol));
+  dimv = INTEGER(getAttrib(V, R_DimSymbol));
+  L = dimu[0]; K = dimv[0];
+  if ((count = dimu[1]) != dimv[1])
+    error("Both 'U' and 'V' should have equal number of columns");
+  N = K + L - 1;
+
+  /* Grab needed data */
+  plan = R_ExternalPtrAddr(fftplan);
+
+  /* Allocate buffer for output */
+  PROTECT(F = allocMatrix(REALSXP, N, count));
+  rF = REAL(F);
+
+  /* Perform the actual hankelization */
+  for (i = 0; i < count; ++i) {
+    R_CheckUserInterrupt();
+    /* TODO: nice target for OpenMP stuff */
+    hankelize_fft(rF+i*N, rU+i*L, rV+i*K, L, K, plan);
+  }
+
+  UNPROTECT(1);
+  return F;
+}
+
 SEXP hankelize_multi(SEXP U, SEXP V) {
   double *rU = REAL(U), *rV = REAL(V), *rF;
   R_len_t L, K, N, i, count;
@@ -686,3 +895,25 @@ SEXP hankelize_multi(SEXP U, SEXP V) {
   UNPROTECT(1);
   return F;
 }
+
+SEXP Lcov_matrix(SEXP F, SEXP L, SEXP fft_plan) {
+  R_len_t intL = INTEGER(L)[0];
+  R_len_t i, j;
+  R_len_t K = length(F) - intL + 1;
+  SEXP ans;
+
+  PROTECT(ans = allocMatrix(REALSXP, intL, intL));
+  double *rans = REAL(ans);
+  double *pF = REAL(F);
+  compute_L_covariation_matrix_first_row(REAL(F), length(F), intL, rans, R_ExternalPtrAddr(fft_plan));
+
+  for (j = 1; j < intL; ++j)
+    rans[intL*j] = rans[j];
+
+  for (j = 1; j < intL; ++j)
+    for (i = j; i < intL; ++i)
+      rans[i + intL * j] = rans[j + intL * i] = rans[(i - 1) + (j - 1) * intL] - pF[i - 1] * pF[j - 1] + pF[i + K - 1] * pF[j + K - 1];
+
+  UNPROTECT(1);
+  return ans;
+}
